

class EndPointParam:
    NO_VALUE = (None, )  # Simple tuple for testing with `is`

    def __init__(self, vtype=str, default=NO_VALUE, help_text=None, required=False):
        self.name = 'param'
        self.data_type = vtype
        self.default = default
        self.help_text = help_text
        self.required = required

    def clean(self, value):
        if value is EndPointParam.NO_VALUE or (value is None and self.required):
            value = self.default

        return value


class ChoicesParam(EndPointParam):
    def __init__(self, choices, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.choices = choices

    def clean(self, value):
        value = super().clean(value)

        if value is not None and value not in self.choices:
            raise ValueError("Invalid Choice. Choices are {}".format(
                ', '.join(self.choices))
            )

        return value


class EndpointMeta(type):
    def __new__(mcs, clsname, bases, clsdict):
        params = []
        for name, val in clsdict.items():
            if isinstance(val, EndPointParam):
                clsdict[name].name = name
                params.append(val)

        clsdict['params'] = tuple(params)
        clsobj = super().__new__(mcs, clsname, bases, clsdict)
        return clsobj


class EndPoint(metaclass=EndpointMeta):
    params = ()

    def __init__(self, uri, params=None):
        self.uri = uri
        self.simple_params = params

    def prepare(self, **kwargs):
        uri = self.uri.format(**kwargs)
        params = {}

        for param in self.params:
            value = param.clean(kwargs.get(param.name, EndPointParam.NO_VALUE))
            if value is not EndPointParam.NO_VALUE:
                params[param.name] = value

        if isinstance(self.simple_params, list):
            for param in self.simple_params:
                value = kwargs.get(param, None)
                if value is not None:
                    params[param] = value

        return uri, params

    def explain(self):
        return "URI: {uri}\n\nDescription: {doc}\n\nParams:\n{params}".format(
            uri=self.uri,
            doc=self.__doc__,
            params="\n".join(
                [
                    '{name}: {help_text}'.format(
                        name=param.name,
                        help_text=param.help_text
                    )
                    for param in self.params
                ]
            )
        )



import inspect

from shutterstock import resources, resource


class Client:
    def __init__(self):
        self.resources = []

    def __setattr__(self, key, value):
        if callable(value) and issubclass(value, resource.Resource):
            self.resources.append(value)
        super().__setattr__(key, value)

    def configure(self):
        for res in self.resources:
            for cls in res.__mro__:
                if issubclass(cls, resources.Resource):
                    for key, var in cls.__dict__.items():
                        if isinstance(var, resource.ResourceMethodAccessor):
                            setattr(res, key, var.configure_for_client(self))


def configure_api(api):
    client = Client()

    for name, val in resources.__dict__.items():
        if inspect.isclass(val) and issubclass(val, resources.Resource):
            setattr(client, name, type(name, (val,), {'API': api}))

    client.configure()
    return client


def configure(token):
    from shutterstock.api import ShutterstockAPI
    return configure_api(ShutterstockAPI(token))



import json
import requests


class ShutterstockAPI:
    def __init__(self, token):
        self.token = token

    @property
    def headers(self):
        return {
            'Content-Type': 'application/json',
            'Authorization': 'Bearer {token}'.format(
                token=self.token
            )
        }

    def request(self, method, endpoint, data_key='params', **params):
        endpoint, params = endpoint.prepare(**params)
        data = {
            data_key: params
        }

        response = method(
            'https://api.shutterstock.com/v2{endpoint}'.format(
                endpoint=endpoint
            ),
            headers=self.headers,
            **data
        )
        print(params)
        print(response.content)
        return json.loads(response.content)

    def get(self, endpoint, **params):
        return self.request(requests.get, endpoint, **params)

    def post(self, endpoint, **params):
        return self.request(requests.post, endpoint, data_key='json', **params)

    def put(self, endpoint, **params):
        return self.request(requests.put, endpoint, data_key='json', **params)

    def delete(self, endpoint, **params):
        return self.request(requests.delete, endpoint, **params)



from functools import partial


class ResourceObjectInitializer:
    def __init__(self, cls, func, instance):
        self.func = func
        self.instance = instance or cls()

    def __call__(self, *args, **kwargs):
        self.instance.reset()
        self.instance.set_data(**self.func(*args, **kwargs))
        return self.instance


class ResourceCollectionInitializer:
    def __init__(self, cls, func, instance):
        self.cls = cls
        self.func = func

    def __call__(self, *args, **kwargs):
        collection = []
        data = self.func(*args, **kwargs)
        if 'errors' in data:
            raise Exception(data['errors'][0]['message'])

        for item in data['data']:
            collection.append(self.cls(**item))
        return collection


class ResourceMethodAccessor:
    def __init__(self, func, initializer, resource=None, **params):
        self.func = func
        self.initializer = initializer
        self.resource = resource
        self.params = params

    def configure_for_client(self, client):
        if self.resource is None:
            return self
        resource = getattr(client, self.resource.__name__)
        return ResourceMethodAccessor(self.func, self.initializer, resource, **self.params)

    def __get__(self, instance, cls):
        params = {}
        if instance:
            for key, val in self.params.items():
                print(key, val)
                params[key] = getattr(instance, val)

        return self.initializer(self.resource or cls, partial(self.func, cls=cls, **params), instance)


class ResourceMethod:
    def __init__(self, initializer, resource=None, **params):
        self.initializer = initializer
        self.resource = resource
        self.params = params

    def __call__(self, func):
        return ResourceMethodAccessor(func, self.initializer, resource=self.resource, **self.params)


class ResourceObjectMethod(ResourceMethod):
    def __init__(self, **params):
        super().__init__(ResourceObjectInitializer, **params)


class ResourceCollectionMethod(ResourceMethod):
    def __init__(self, **params):
        super().__init__(ResourceCollectionInitializer, **params)


class Resource:
    API = None
    LIST = None
    GET = None
    CREATE = None
    UPDATE = None
    DELETE = None

    def __init__(self, **kwargs):
        self.set_data(**kwargs)

    def reset(self):
        pass

    def set_data(self, **params):
        for key, val in params.items():
            setattr(self, key, val)

    @ResourceCollectionMethod()
    def all(cls, **params):
        return cls.API.get(cls.LIST, **params)

    @ResourceCollectionMethod()
    def list(cls, **params):
        return cls.API.get(cls.LIST, **params)

    @ResourceObjectMethod(id='id')
    def get(cls, **params):
        return cls.API.get(cls.GET, **params)


from shutterstock.endpoint import EndPoint, EndPointParam, ChoicesParam
from shutterstock.resource import Resource, ResourceObjectMethod, \
    ResourceCollectionMethod


class ImageEndPoint(EndPoint):
    """Endpoint for Shutterstock images"""
    MINIMAL = 'minimal'
    FULL = 'full'
    VIEW_CHOICES = (MINIMAL, FULL, )

    id = EndPointParam(required=True,
                       help_text='Required. The ID of the image.')
    view = ChoicesParam(required=True, default=MINIMAL, choices=VIEW_CHOICES,
                        help_text='Required. Minimal view does not return licensing options, categories, keywords')


class Contributor(Resource):
    LIST = EndPoint('/contributors')
    GET = EndPoint('/contributors/{id}')


class Image(Resource):
    LIST = ImageEndPoint('/images')
    GET = ImageEndPoint('/images/{id}')


class ImageCollectionListEndPoint(EndPoint):
    id = EndPointParam()


class ImageCollection(Resource):
    LIST = ImageCollectionListEndPoint('/images/collections')
    GET = EndPoint('/images/collections/{id}')
    ITEMS = EndPoint('/images/collections/{id}/items')

    @ResourceCollectionMethod(resource=Image, id='id')
    def items(cls, **params):
        response = cls.API.get(cls.ITEMS, **params)
        ids = [item['id'] for item in response['data']]
        return cls.API.get(Image.LIST, id=ids, view=params.get('view', 'minimal'))


class ImageLicense(Resource):
    LIST = EndPoint('/images/licenses')
    DOWNLOAD = EndPoint('/images/licenses/{id}/downloads')
    LICENSE = EndPoint('/images/licenses?subscription_id={subscription_id}', params=['images'])

    @ResourceObjectMethod(id='id')
    def download(cls, **params):
        return cls.API.post(cls.DOWNLOAD, **params)

    @ResourceCollectionMethod(id='id')
    def license(cls, **params):
        return cls.API.post(cls.LICENSE, **params)
    
    
class ImageContributor(Resource):
    GET = EndPoint('/contributors/{contributor_id}')
